use std::io::{self, stdin, Read};

// use minifb::{InputCallback, Key};
use termion::{
    event::Key,
    input::{Keys, TermRead},
};

use crate::memory::{MemHandle, MemoryWriter};

pub struct Keyboard {
    memory: MemHandle,
}

struct PlainKeys<R> {
    iter: Keys<R>,
}

impl<R: Read> Iterator for PlainKeys<R> {
    type Item = Result<char, io::Error>;
    fn next(&mut self) -> Option<Self::Item> {
        loop {
            match self.iter.next() {
                Some(Ok(key)) => match key {
                    Key::Ctrl(char) => return Some(Ok(char)),
                    Key::Alt(char) => return Some(Ok(char)),
                    Key::Char('A') => return Some(Ok('a')),
                    Key::Char('B') => return Some(Ok('b')),
                    Key::Char('C') => return Some(Ok('c')),
                    Key::Char('D') => return Some(Ok('d')),
                    Key::Char('E') => return Some(Ok('e')),
                    Key::Char('F') => return Some(Ok('f')),
                    Key::Char('G') => return Some(Ok('g')),
                    Key::Char('H') => return Some(Ok('h')),
                    Key::Char('I') => return Some(Ok('i')),
                    Key::Char('J') => return Some(Ok('j')),
                    Key::Char('K') => return Some(Ok('k')),
                    Key::Char('L') => return Some(Ok('l')),
                    Key::Char('M') => return Some(Ok('m')),
                    Key::Char('N') => return Some(Ok('n')),
                    Key::Char('O') => return Some(Ok('o')),
                    Key::Char('P') => return Some(Ok('p')),
                    Key::Char('Q') => return Some(Ok('q')),
                    Key::Char('R') => return Some(Ok('r')),
                    Key::Char('S') => return Some(Ok('s')),
                    Key::Char('T') => return Some(Ok('t')),
                    Key::Char('U') => return Some(Ok('u')),
                    Key::Char('V') => return Some(Ok('v')),
                    Key::Char('W') => return Some(Ok('w')),
                    Key::Char('X') => return Some(Ok('x')),
                    Key::Char('Y') => return Some(Ok('y')),
                    Key::Char('Z') => return Some(Ok('z')),
                    Key::Char('!') => return Some(Ok('1')),
                    Key::Char('@') => return Some(Ok('2')),
                    Key::Char('#') => return Some(Ok('3')),
                    Key::Char('$') => return Some(Ok('4')),
                    Key::Char('%') => return Some(Ok('5')),
                    Key::Char('^') => return Some(Ok('6')),
                    Key::Char('&') => return Some(Ok('7')),
                    Key::Char('*') => return Some(Ok('8')),
                    Key::Char('(') => return Some(Ok('9')),
                    Key::Char(')') => return Some(Ok('0')),
                    Key::Char('~') => return Some(Ok('`')),
                    Key::Char('_') => return Some(Ok('-')),
                    Key::Char('+') => return Some(Ok('=')),
                    Key::Char('|') => return Some(Ok('\\')),
                    Key::Char('}') => return Some(Ok(']')),
                    Key::Char('{') => return Some(Ok('[')),
                    Key::Char('"') => return Some(Ok('\'')),
                    Key::Char(':') => return Some(Ok(';')),
                    Key::Char('?') => return Some(Ok('/')),
                    Key::Char('>') => return Some(Ok('.')),
                    Key::Char('<') => return Some(Ok(',')),
                    Key::Char(char) => return Some(Ok(char)),
                    _ => continue,
                },
                Some(Err(e)) => return Some(Err(e)),
                None => return None,
            }
        }
    }
}

#[derive(Clone, Copy, Debug, Ord, Eq, PartialEq, PartialOrd)]
enum ControlKey {
    Alt,
    Shift,
    Ctrl,
    Left,
    Right,
    Up,
    Down,
    Esc,
    Enter,
    Home,
    End,
    Tab,
    Backspace,
    Delete,
    Insert,
    PageDown,
    PageUp,
}

struct ControlKeys<R> {
    iter: Keys<R>,
    extra: Option<ControlKey>,
}

impl<R: Read> Iterator for ControlKeys<R> {
    type Item = Result<ControlKey, io::Error>;
    fn next(&mut self) -> Option<Self::Item> {
        loop {
            if let Some(key) = self.extra {
                return Some(Ok(key));
            }
            match self.iter.next() {
                Some(Ok(key)) => match key {
                    Key::Ctrl(char) => match char {
                        'A' | 'B' | 'C' | 'D' | 'E' | 'F' | 'G' | 'H' | 'I' | 'J' | 'K' | 'L'
                        | 'M' | 'N' | 'O' | 'P' | 'Q' | 'R' | 'S' | 'T' | 'U' | 'V' | 'W' | 'X'
                        | 'Y' | 'Z' | '!' | '@' | '#' | '$' | '%' | '^' | '&' | '*' | '(' | ')'
                        | '~' | '_' | '+' | '|' | '}' | '{' | '"' | ':' | '?' | '>' | '<' => {
                            self.extra = Some(ControlKey::Shift);
                            return Some(Ok(ControlKey::Ctrl));
                        }
                        _ => return Some(Ok(ControlKey::Ctrl)),
                    },
                    Key::CtrlUp => {
                        self.extra = Some(ControlKey::Ctrl);
                        return Some(Ok(ControlKey::Up));
                    }
                    Key::CtrlDown => {
                        self.extra = Some(ControlKey::Ctrl);
                        return Some(Ok(ControlKey::Down));
                    }
                    Key::CtrlRight => {
                        self.extra = Some(ControlKey::Ctrl);
                        return Some(Ok(ControlKey::Right));
                    }
                    Key::CtrlLeft => {
                        self.extra = Some(ControlKey::Ctrl);
                        return Some(Ok(ControlKey::Left));
                    }
                    Key::CtrlEnd => {
                        self.extra = Some(ControlKey::Ctrl);
                        return Some(Ok(ControlKey::End));
                    }
                    Key::CtrlHome => {
                        self.extra = Some(ControlKey::Ctrl);
                        return Some(Ok(ControlKey::Home));
                    }
                    Key::Alt(char) => match char {
                        'A' | 'B' | 'C' | 'D' | 'E' | 'F' | 'G' | 'H' | 'I' | 'J' | 'K' | 'L'
                        | 'M' | 'N' | 'O' | 'P' | 'Q' | 'R' | 'S' | 'T' | 'U' | 'V' | 'W' | 'X'
                        | 'Y' | 'Z' | '!' | '@' | '#' | '$' | '%' | '^' | '&' | '*' | '(' | ')'
                        | '~' | '_' | '+' | '|' | '}' | '{' | '"' | ':' | '?' | '>' | '<' => {
                            self.extra = Some(ControlKey::Shift);
                            return Some(Ok(ControlKey::Alt));
                        }
                        _ => return Some(Ok(ControlKey::Alt)),
                    },
                    Key::AltUp => {
                        self.extra = Some(ControlKey::Alt);
                        return Some(Ok(ControlKey::Up));
                    }
                    Key::AltDown => {
                        self.extra = Some(ControlKey::Alt);
                        return Some(Ok(ControlKey::Down));
                    }
                    Key::AltRight => {
                        self.extra = Some(ControlKey::Alt);
                        return Some(Ok(ControlKey::Right));
                    }
                    Key::AltLeft => {
                        self.extra = Some(ControlKey::Alt);
                        return Some(Ok(ControlKey::Left));
                    }
                    Key::Char('A')
                    | Key::Char('B')
                    | Key::Char('C')
                    | Key::Char('D')
                    | Key::Char('E')
                    | Key::Char('F')
                    | Key::Char('G')
                    | Key::Char('H')
                    | Key::Char('I')
                    | Key::Char('J')
                    | Key::Char('K')
                    | Key::Char('L')
                    | Key::Char('M')
                    | Key::Char('N')
                    | Key::Char('O')
                    | Key::Char('P')
                    | Key::Char('Q')
                    | Key::Char('R')
                    | Key::Char('S')
                    | Key::Char('T')
                    | Key::Char('U')
                    | Key::Char('V')
                    | Key::Char('W')
                    | Key::Char('X')
                    | Key::Char('Y')
                    | Key::Char('Z')
                    | Key::Char('!')
                    | Key::Char('@')
                    | Key::Char('#')
                    | Key::Char('$')
                    | Key::Char('%')
                    | Key::Char('^')
                    | Key::Char('&')
                    | Key::Char('*')
                    | Key::Char('(')
                    | Key::Char(')')
                    | Key::Char('~')
                    | Key::Char('_')
                    | Key::Char('+')
                    | Key::Char('|')
                    | Key::Char('}')
                    | Key::Char('{')
                    | Key::Char('"')
                    | Key::Char(':')
                    | Key::Char('?')
                    | Key::Char('>')
                    | Key::Char('<') => return Some(Ok(ControlKey::Shift)),
                    Key::ShiftUp => {
                        self.extra = Some(ControlKey::Shift);
                        return Some(Ok(ControlKey::Up));
                    }
                    Key::ShiftDown => {
                        self.extra = Some(ControlKey::Shift);
                        return Some(Ok(ControlKey::Down));
                    }
                    Key::ShiftRight => {
                        self.extra = Some(ControlKey::Shift);
                        return Some(Ok(ControlKey::Right));
                    }
                    Key::ShiftLeft => {
                        self.extra = Some(ControlKey::Shift);
                        return Some(Ok(ControlKey::Left));
                    }
                    Key::Backspace => return Some(Ok(ControlKey::Backspace)),
                    Key::Up => return Some(Ok(ControlKey::Up)),
                    Key::Down => return Some(Ok(ControlKey::Down)),
                    Key::Right => return Some(Ok(ControlKey::Right)),
                    Key::Left => return Some(Ok(ControlKey::Left)),
                    Key::Home => return Some(Ok(ControlKey::Home)),
                    Key::End => return Some(Ok(ControlKey::End)),
                    Key::PageUp => return Some(Ok(ControlKey::PageUp)),
                    Key::PageDown => return Some(Ok(ControlKey::PageDown)),
                    Key::Insert => return Some(Ok(ControlKey::Insert)),
                    Key::Delete => return Some(Ok(ControlKey::Delete)),
                    _ => continue,
                },
                Some(Err(e)) => return Some(Err(e)),
                None => return None,
            }
        }
    }
}

trait GetKeys {
    fn get_plain(self) -> PlainKeys<Self>
    where
        Self: Sized;
    fn get_modifiers(self) -> ControlKeys<Self>
    where
        Self: Sized;
}

impl<R: Read + TermRead> GetKeys for R {
    fn get_plain(self) -> PlainKeys<Self>
    where
        Self: Sized,
    {
        PlainKeys { iter: self.keys() }
    }
    fn get_modifiers(self) -> ControlKeys<Self>
    where
        Self: Sized,
    {
        ControlKeys {
            iter: self.keys(),
            extra: None,
        }
    }
}

impl Keyboard {
    pub fn new(memory: MemHandle) -> Self {
        Self { memory }
    }

    pub fn clear_keys(&self) {
        self.memory.write(0x4400, 0x00);
        self.memory.write(0x4401, 0x00);
        self.memory.write(0x4402, 0x00);
        self.memory.write(0x4403, 0x00);
        self.memory.write(0x4404, 0x00);
        self.memory.write(0x4405, 0x00);
    }

    // pub fn read_keys(&self, key: KeyEvent) {
    pub fn read_keys(&self) {
        let mut row0 = 0;
        let mut row1 = 0;
        let mut row2 = 0;
        let mut row3 = 0;
        let mut row4 = 0;
        let mut row5 = 0;

        let modifiers = termion::async_stdin().get_modifiers();
        for m in modifiers {
            match m.unwrap() {
                ControlKey::Shift => row2 ^= 0b1000_0000,
                ControlKey::Ctrl => row3 ^= 0b1000_0000,
                ControlKey::Alt => row4 ^= 0b1000_0000,
                ControlKey::Esc => row0 ^= 0b1000_0000,
                ControlKey::Backspace => row0 ^= 0b0000_0001,
                ControlKey::Tab => row1 ^= 0b1000_0000,
                ControlKey::Enter => row1 ^= 0b0000_0001,
                _ => {}
            }
        }

        let keys = termion::async_stdin().get_plain();
        for k in keys {
            match k.unwrap() {
                'w' => row0 ^= 0b0100_0000,
                'e' => row0 ^= 0b0010_0000,
                'r' => row0 ^= 0b0001_0000,
                't' => row0 ^= 0b0000_1000,
                'u' => row0 ^= 0b0000_0100,
                'o' => row0 ^= 0b0000_0010,
                'q' => row1 ^= 0b0100_0000,
                's' => row1 ^= 0b0010_0000,
                'g' => row1 ^= 0b0001_0000,
                'y' => row1 ^= 0b0000_1000,
                'i' => row1 ^= 0b0000_0100,
                'p' => row1 ^= 0b0000_0010,
                'd' => row2 ^= 0b0100_0000,
                'v' => row2 ^= 0b0010_0000,
                'h' => row2 ^= 0b0001_0000,
                'k' => row2 ^= 0b0000_1000,
                '\\' => row2 ^= 0b0000_0100,
                '/' => row2 ^= 0b0000_0010,
                'a' => row2 ^= 0b0000_0001,
                'z' => row3 ^= 0b0100_0000,
                'f' => row3 ^= 0b0010_0000,
                'b' => row3 ^= 0b0001_0000,
                'j' => row3 ^= 0b0000_1000,
                'l' => row3 ^= 0b0000_0100,
                '2' => row3 ^= 0b0000_0010,
                '4' => row3 ^= 0b0000_0001,
                'x' => row4 ^= 0b0100_0000,
                'c' => row4 ^= 0b0010_0000,
                'n' => row4 ^= 0b0001_0000,
                'm' => row4 ^= 0b0000_1000,
                ',' => row4 ^= 0b0000_0100,
                '1' => row4 ^= 0b0000_0010,
                '3' => row4 ^= 0b0000_0001,
                ' ' => row5 ^= 0b0100_0000,
                _ => {}
            }
        }
        // match key.code {
        //     KeyCode::Esc => row0 ^= 0b1000_0000,
        //     KeyCode::Char('w') => row0 ^= 0b0100_0000,
        //     KeyCode::Char('e') => row0 ^= 0b0010_0000,
        //     KeyCode::Char('r') => row0 ^= 0b0001_0000,
        //     KeyCode::Char('t') => row0 ^= 0b0000_1000,
        //     KeyCode::Char('u') => row0 ^= 0b0000_0100,
        //     KeyCode::Char('o') => row0 ^= 0b0000_0010,
        //     KeyCode::Backspace => row0 ^= 0b0000_0001,
        //     KeyCode::Tab => row1 ^= 0b1000_0000,
        //     KeyCode::Char('q') => row1 ^= 0b0100_0000,
        //     KeyCode::Char('s') => row1 ^= 0b0010_0000,
        //     KeyCode::Char('g') => row1 ^= 0b0001_0000,
        //     KeyCode::Char('y') => row1 ^= 0b0000_1000,
        //     KeyCode::Char('i') => row1 ^= 0b0000_0100,
        //     KeyCode::Char('p') => row1 ^= 0b0000_0010,
        //     KeyCode::Enter => row1 ^= 0b0000_0001,
        //     KeyCode::Char('d') => row2 ^= 0b0100_0000,
        //     KeyCode::Char('v') => row2 ^= 0b0010_0000,
        //     KeyCode::Char('h') => row2 ^= 0b0001_0000,
        //     KeyCode::Char('k') => row2 ^= 0b0000_1000,
        //     KeyCode::Char('\'') => row2 ^= 0b0000_0100,
        //     KeyCode::Char('/') => row2 ^= 0b0000_0010,
        //     KeyCode::Char('a') => row2 ^= 0b0000_0001,
        //     KeyCode::Char('z') => row3 ^= 0b0100_0000,
        //     KeyCode::Char('f') => row3 ^= 0b0010_0000,
        //     KeyCode::Char('b') => row3 ^= 0b0001_0000,
        //     KeyCode::Char('j') => row3 ^= 0b0000_1000,
        //     KeyCode::Char('l') => row3 ^= 0b0000_0100,
        //     KeyCode::Char('2') => row3 ^= 0b0000_0010,
        //     KeyCode::Char('4') => row3 ^= 0b0000_0001,
        //     KeyCode::Char('x') => row4 ^= 0b0100_0000,
        //     KeyCode::Char('c') => row4 ^= 0b0010_0000,
        //     KeyCode::Char('n') => row4 ^= 0b0001_0000,
        //     KeyCode::Char('m') => row4 ^= 0b0000_1000,
        //     KeyCode::Char(',') => row4 ^= 0b0000_0100,
        //     KeyCode::Char('1') => row4 ^= 0b0000_0010,
        //     KeyCode::Char('3') => row4 ^= 0b0000_0001,
        //     KeyCode::Char(' ') => row5 ^= 0b0100_0000,
        //     _ => {
        //         row0 = 0;
        //         row1 = 0;
        //         row2 = 0;
        //         row3 = 0;
        //         row4 = 0;
        //         row5 = 0;
        //     }
        // };

        self.memory.write(0x4400, row0);
        self.memory.write(0x4401, row1);
        self.memory.write(0x4402, row2);
        self.memory.write(0x4403, row3);
        self.memory.write(0x4404, row4);
        self.memory.write(0x4405, row5);
    }
}

impl MemoryWriter for Keyboard {
    fn write(&self, address: u16, data: u8) {
        if data != 0x00 {
            println!("wrote {:02x} to address {:04x}", data, address);
        }
        self.memory.write(address, data);
    }
}

// impl InputCallback for Keyboard {
//     fn add_char(&mut self, _uni_char: u32) {}
//     fn set_key_state(&mut self, key: Key, _state: bool) {
//         let mut row0 = 0;
//         let mut row1 = 0;
//         let mut row2 = 0;
//         let mut row3 = 0;
//         let mut row4 = 0;
//         let mut row5 = 0;

//         match key {
//             Key::Escape => row0 ^= 0b1000_0000,
//             Key::W => row0 ^= 0b0100_0000,
//             Key::E => row0 ^= 0b0010_0000,
//             Key::R => row0 ^= 0b0001_0000,
//             Key::T => row0 ^= 0b0000_1000,
//             Key::U => row0 ^= 0b0000_0100,
//             Key::O => row0 ^= 0b0000_0010,
//             Key::Backspace => row0 ^= 0b0000_0001,
//             Key::Tab => row1 ^= 0b1000_0000,
//             Key::Q => row1 ^= 0b0100_0000,
//             Key::S => row1 ^= 0b0010_0000,
//             Key::G => row1 ^= 0b0001_0000,
//             Key::Y => row1 ^= 0b0000_1000,
//             Key::I => row1 ^= 0b0000_0100,
//             Key::P => row1 ^= 0b0000_0010,
//             Key::Enter => row1 ^= 0b0000_0001,
//             Key::LeftShift | Key::RightShift => row2 ^= 0b1000_0000,
//             Key::D => row2 ^= 0b0100_0000,
//             Key::V => row2 ^= 0b0010_0000,
//             Key::H => row2 ^= 0b0001_0000,
//             Key::K => row2 ^= 0b0000_1000,
//             Key::Apostrophe => row2 ^= 0b0000_0100,
//             Key::Slash => row2 ^= 0b0000_0010,
//             Key::A => row2 ^= 0b0000_0001,
//             Key::LeftCtrl | Key::RightCtrl => row3 ^= 0b1000_0000,
//             Key::Z => row3 ^= 0b0100_0000,
//             Key::F => row3 ^= 0b0010_0000,
//             Key::B => row3 ^= 0b0001_0000,
//             Key::J => row3 ^= 0b0000_1000,
//             Key::L => row3 ^= 0b0000_0100,
//             Key::Key2 => row3 ^= 0b0000_0010,
//             Key::Key4 => row3 ^= 0b0000_0001,
//             Key::LeftAlt | Key::RightAlt => row4 ^= 0b1000_0000,
//             Key::X => row4 ^= 0b0100_0000,
//             Key::C => row4 ^= 0b0010_0000,
//             Key::N => row4 ^= 0b0001_0000,
//             Key::M => row4 ^= 0b0000_1000,
//             Key::Comma => row4 ^= 0b0000_0100,
//             Key::Key1 => row4 ^= 0b0000_0010,
//             Key::Key3 => row4 ^= 0b0000_0001,
//             Key::LeftSuper => row5 ^= 0b1000_0000,
//             Key::Space => row5 ^= 0b0100_0000,
//             Key::RightSuper => row5 ^= 0b0010_0000,
//             _ => {}
//         };

//         self.memory.write(0x4400, row0);
//         self.memory.write(0x4401, row1);
//         self.memory.write(0x4402, row2);
//         self.memory.write(0x4403, row3);
//         self.memory.write(0x4404, row4);
//         self.memory.write(0x4405, row5);
//     }
// }
